from math import inf
from typing import List, Dict, Optional
from estimators.bandits.cressieread import EstimatorImpl, IntervalImpl
from estimators.math import IncrementalFsum, clopper_pearson


class Estimator():
    wmin: float
    wmax: float
    n: IncrementalFsum
    _impl: Dict[str, EstimatorImpl]


    def __init__(self, wmin: float = 0, wmax: float = inf):
        self.wmin = wmin
        self.wmax = wmax
        self.n = IncrementalFsum()
        self._impl = {}

    def add_example(self, slot_ids: List[str], p_logs: List[float], rs: List[float], p_preds: List[float]) -> None:
        if len(p_logs) != len(rs) or len(rs) != len(p_preds) or len(p_preds) != len(set(slot_ids)):
            raise ValueError(f'Error: unique elements in slot_ids and length of p_logs, rs, p_preds must be the same, \
                found {len(set(slot_ids))}, {len(p_logs)}, {len(rs)}, {len(p_preds)}  respectively')

        self.n += 1
        ws = [p_pred / p_log for p_pred, p_log in zip(p_preds, p_logs)]
        w = 1.0
        for i in range(len(ws)):
            w *= ws[i]
            if slot_ids[i] not in self._impl:
                self._impl[slot_ids[i]] = EstimatorImpl(0, inf)
            self._impl[slot_ids[i]].add(w, rs[i])

    def get_impression(self) -> Dict[str, float]:
        result = {}
        if float(self.n) > 0:
            for slot_id, estimator in self._impl.items():
                result[slot_id] = float(estimator.n) / float(self.n)
        return result

    def get_r_given_impression(self) -> Dict[str, float]:
        result = {}
        if float(self.n) > 0:
            for slot_id, estimator in self._impl.items():
                result[slot_id] = estimator.get()
        return result

    def get_r(self) -> Dict[str, float]:
        result = {}
        if float(self.n) > 0:
            impression = self.get_impression()
            r_given_impression = self.get_r_given_impression()
            for slot_id, estimator in self._impl.items():
                result[slot_id] = impression[slot_id] * r_given_impression[slot_id]
        return result

    def get_r_overall(self) -> float:
        return sum(self._impl.values(), EstimatorImpl(0, inf)).get()

    def __add__(self, other: 'Estimator') -> 'Estimator':
        slot_ids = set(self._impl.keys()).union(set(other._impl.keys()))
        result = Estimator(wmin = min(self.wmin, other.wmin), wmax = max(self.wmax, other.wmax))
        result.n = self.n + other.n
        default = lambda: EstimatorImpl(0, inf)
        for id in slot_ids:
            result._impl[id] = self._impl.get(id, default()) + other._impl.get(id, default())
        return result


class Interval():
    rmin: float
    rmax: float
    n: IncrementalFsum
    _impl: Dict[str, IntervalImpl]

    def __init__(self, rmin: float = 0, rmax: float = 1, empirical_r_bounds: bool = False):
        self.rmin = rmin
        self.rmax = rmax
        self.n = IncrementalFsum()
        self._impl = {}
        self.empirical_r_bounds = empirical_r_bounds

    def add_example(self, slot_ids: List[str], p_logs: List[float], rs: List[float], p_preds: List[float], p_drop: float = 0, n_drop: Optional[int] = None) -> None:
        if len(p_logs) != len(rs) or len(rs) != len(p_preds) or len(p_preds) != len(set(slot_ids)):
            raise ValueError(f'Error: unique elements in slot_ids and length of p_logs, rs, p_preds must be the same, \
                found {len(set(slot_ids))}, {len(p_logs)}, {len(rs)}, {len(p_preds)}  respectively')

        self.n += 1
        ws = [p_pred / p_log for p_pred, p_log in zip(p_preds, p_logs)]
        w = 1.0
        for i in range(len(ws)):
            w *= ws[i]
            if slot_ids[i] not in self._impl:
                self._impl[slot_ids[i]] = IntervalImpl(0, inf, self.rmin, self.rmax, self.empirical_r_bounds)
            self._impl[slot_ids[i]].add(w, rs[i], p_drop, n_drop)

    def get_impression(self, alpha: float = 0.05) -> Dict[str, List[float]]:
        result = {}
        if float(self.n) > 0:
            for slot_id, estimator in self._impl.items():
                result[slot_id] = clopper_pearson(float(estimator.n), float(self.n), alpha)
        return result

    def get_r_given_impression(self, alpha: float = 0.05, atol: float = 1e-9) -> Dict[str, List[float]]:
        result = {}
        if float(self.n) > 0:
            for slot_id, estimator in self._impl.items():
                result[slot_id] = estimator.get(alpha, atol)
        return result

    def get_r(self, alpha: float = 0.05, atol: float = 1e-9) -> Dict[str, List[float]]:
        result = {}
        if float(self.n) > 0:
            impression = self.get_impression(alpha)
            r_given_impression = self.get_r_given_impression(alpha, atol)
            for slot_id, estimator in self._impl.items():
                result[slot_id] = [a * b for a, b in zip(impression[slot_id], r_given_impression[slot_id])]
        return result

    def get_r_overall(self, alpha: float = 0.05, atol: float = 1e-9) -> List[float]:
        return sum(self._impl.values(), IntervalImpl(0, inf, self.rmin, self.rmax, self.empirical_r_bounds)).get(alpha, atol)

    def __add__(self, other: 'Interval') -> 'Interval':
        assert not (self.empirical_r_bounds ^ other.empirical_r_bounds), 'Summation of estimators with various r bounds policy is prohibited'
        
        if not self.empirical_r_bounds:
            assert self.rmin == other.rmin, 'Summation of estimators with various r bounds is prohibited'
            assert self.rmax == other.rmax, 'Summation of estimators with various r bounds is prohibited'
        rmin = min(self.rmin, other.rmin)
        rmax = max(self.rmax, other.rmax)
        slot_ids = set(self._impl.keys()).union(set(other._impl.keys()))
        result = Interval(rmin=rmin, rmax=rmax, empirical_r_bounds=self.empirical_r_bounds)
        result.n = self.n + other.n
        default = lambda: IntervalImpl(wmin=0, wmax=inf, rmin=rmin, rmax=rmax, empirical_r_bounds=self.empirical_r_bounds)
        for id in slot_ids:
            result._impl[id] = self._impl.get(id, default()) + other._impl.get(id, default())
        return result



import math
from estimators.slates import base
from scipy import stats
from typing import List, Optional


class Interval(base.Interval):
    examples_count: float
    weighted_reward: float
    weighted_reward_sq: float

    def __init__(self):
        self.examples_count = 0
        self.weighted_reward = 0
        self.weighted_reward_sq = 0

    def add_example(self, p_logs: List[float], r: float, p_preds: List[float], count: float = 1.0) -> None:
        """Expects lists for logged probabilities and predicted probabilities. These should correspond to each slot.
        This function is implemented under the simplifying assumptions of
        example 4 in the paper 'Off-policy evaluation for slate recommendation'
        where the slate space is a cartesian product and the logging policy is a
        product distribution"""
        if not isinstance(p_logs, list) or not isinstance(p_preds, list):
            raise ValueError('Error: p_logs and p_preds must be lists')

        if len(p_logs) != len(p_preds):
            raise ValueError(f'Error: p_logs and p_preds must be the same length, found {len(p_logs)} '
                             f'and {len(p_preds)} respectively')

        self.examples_count += count
        num_slots = len(p_logs)
        w = 1 - num_slots
        for p_log, p_pred in zip(p_logs, p_preds):
            w += p_pred / p_log
        self.weighted_reward += r * w * count
        self.weighted_reward_sq += ((r * w) ** 2) * count

    def get(self, alpha: float = 0.05) -> List[Optional[float]]:
        if self.examples_count <= 1:
            return [None, None]

        z_gaussian_cdf = stats.norm.ppf(1 - alpha / 2)
        variance = (self.weighted_reward_sq - self.weighted_reward * self.weighted_reward / self.examples_count) \
            / (self.examples_count - 1)
        gauss_delta = z_gaussian_cdf * math.sqrt(variance / self.examples_count)
        ips = self.weighted_reward / self.examples_count
        return [ips - gauss_delta, ips + gauss_delta]

    def __add__(self, other: 'Interval') -> 'Interval':
        result = Interval()
        result.examples_count = self.examples_count + other.examples_count
        result.weighted_reward = self.weighted_reward + other.weighted_reward
        result.weighted_reward_sq = self.weighted_reward_sq + other.weighted_reward_sq
        return result



from estimators.slates import base
from typing import List, Optional

# PseudoInverse estimator for slate recommendation. The following implements the
# case for a Cartesian product when mu is a product distribution. This can be
# seen in example 4 of the paper.
# https://arxiv.org/abs/1605.04812


class Estimator(base.Estimator):
    examples_count: float
    weighted_reward: float

    def __init__(self):
        self.examples_count = 0
        self.weighted_reward = 0

    def add_example(self, p_logs: List[float], r: float, p_preds: List[float], count: float = 1.0) -> None:
        """Expects lists for logged probabilities and predicted probabilities. These should correspond to each slot.
        This function is implemented under the simplifying assumptions of
        example 4 in the paper 'Off-policy evaluation for slate recommendation'
        where the slate space is a cartesian product and the logging policy is a
        product distribution"""
        if not isinstance(p_logs, list) or not isinstance(p_preds, list):
            raise ValueError('Error: p_logs and p_preds must be lists')

        if len(p_logs) != len(p_preds):
            raise ValueError(f'Error: p_logs and p_preds must be the same length, found {len(p_logs)} '
                             f'and {len(p_preds)} respectively')

        self.examples_count += count
        num_slots = len(p_logs)
        w = 1 - num_slots
        for p_log, p_pred in zip(p_logs, p_preds):
            w += p_pred/p_log
        self.weighted_reward += r * w * count

    def get(self) -> Optional[float]:
        if self.examples_count > 0:
            return self.weighted_reward / self.examples_count
        return None

    def __add__(self, other: 'Estimator') -> 'Estimator':
        result = Estimator()
        result.examples_count = self.examples_count + other.examples_count
        result.weighted_reward = self.weighted_reward + other.weighted_reward
        return result



import argparse, os, gzip
from bandits import cressieread
from bandits import ips
from bandits import snips
from bandits import mle
from bandits import gaussian
from bandits import clopper_pearson
from bandits import cats_utils
from utils import ds_parse


def compute_estimates(log_fp, cats_transformer=None):
    # Init estimators
    online_ips = ips.Estimator()
    online_snips = snips.Estimator()
    online_mle = mle.Estimator()
    online_cressieread = cressieread.Estimator()

    baseline1_ips = ips.Estimator()
    baseline1_snips = snips.Estimator()
    baseline1_mle = mle.Estimator()
    baseline1_cressieread = cressieread.Estimator()

    baselineR_ips = ips.Estimator()
    baselineR_snips = snips.Estimator()
    baselineR_mle = mle.Estimator()
    baselineR_cressieread = cressieread.Estimator()

    baseline1_gaussian = gaussian.Interval()
    baseline1_clopper_pearson = clopper_pearson.Interval()

    baselineR_gaussian = gaussian.Interval()
    baselineR_clopper_pearson = clopper_pearson.Interval()

    print('Processing: {}'.format(log_fp))
    bytes_count = 0
    tot_bytes = os.path.getsize(log_fp)
    evts = 0
    for i,x in enumerate(gzip.open(log_fp, 'rb') if log_fp.endswith('.gz') else open(log_fp, 'rb')):
        # display progress
        bytes_count += len(x)
        if (i+1) % 10000 == 0:
            if log_fp.endswith('.gz'):
                ds_parse.update_progress(i+1)
            else:
                ds_parse.update_progress(bytes_count,tot_bytes)

        # parse dsjson file
        if x.startswith(b'{"_label_cost":') and x.strip().endswith(b'}'):
            data = ds_parse.json_cooked(x)

            if data['skipLearn']:
                continue

            r = 0 if data['cost'] == b'0' else -float(data['cost'])

            # Update estimators with tuple (p_log, r, p_pred)
            online_ips.add_example(data['p'], r, data['p'])
            baseline1_ips.add_example(data['p'], r, 1 if data['a'] == 1 else 0)
            baselineR_ips.add_example(data['p'], r, 1/data['num_a'])

            online_snips.add_example(data['p'], r, data['p'])
            baseline1_snips.add_example(data['p'], r, 1 if data['a'] == 1 else 0)
            baselineR_snips.add_example(data['p'], r, 1/data['num_a'])

            online_mle.add_example(data['p'], r, data['p'])
            baseline1_mle.add_example(data['p'], r, 1 if data['a'] == 1 else 0)
            baselineR_mle.add_example(data['p'], r, 1/data['num_a'])

            online_cressieread.add_example(data['p'], r, data['p'])
            baseline1_cressieread.add_example(data['p'], r, 1 if data['a'] == 1 else 0)
            baselineR_cressieread.add_example(data['p'], r, 1/data['num_a'])

            baseline1_gaussian.add_example(data['p'], r, 1 if data['a'] == 1 else 0)
            baseline1_clopper_pearson.add_example(data['p'], r, 1 if data['a'] == 1 else 0)

            baselineR_gaussian.add_example(data['p'], r, 1/data['num_a'])
            baselineR_clopper_pearson.add_example(data['p'], r, 1/data['num_a'])

            evts += 1

        if x.startswith(b'{"_label_ca":') and x.strip().endswith(b'}'):
            data = ds_parse.json_cooked_continuous_actions(x)
            if cats_transformer is None:
                raise RuntimeError("Not all of the required arguments for running with continuous actions have been provided.")
            # passing logged action as predicted action to transformer
            data = cats_transformer.transform(data, data['a'])
            # passing baseline action as predicted action to transformer
            data_baseline1 = cats_transformer.transform(data, cats_transformer.get_baseline1_prediction())

            if data['skipLearn']:
                continue

            r = 0 if data['cost'] == b'0' else -float(data['cost'])

            # Update estimators with tuple (p_log, r, p_pred)
            online_ips.add_example(data['p'], r, data['p'])
            baseline1_ips.add_example(data['p'], r, data_baseline1['pred_p'])
            baselineR_ips.add_example(data['p'], r, 1.0 / cats_transformer.continuous_range)

            online_snips.add_example(data['p'], r, data['p'])
            baseline1_snips.add_example(data['p'], r, data_baseline1['pred_p'])
            baselineR_snips.add_example(data['p'], r, 1.0 / cats_transformer.continuous_range)

            online_mle.add_example(data['p'], r, data['p'])
            baseline1_mle.add_example(data['p'], r, data_baseline1['pred_p'])
            baselineR_mle.add_example(data['p'], r, 1.0 / cats_transformer.continuous_range)

            online_cressieread.add_example(data['p'], r, data['p'])
            baseline1_cressieread.add_example(data['p'], r, data_baseline1['pred_p'])
            baselineR_cressieread.add_example(data['p'], r, 1.0 / cats_transformer.continuous_range)

            baseline1_gaussian.add_example(data['p'], r, data_baseline1['pred_p'])
            baseline1_clopper_pearson.add_example(data['p'], r, data_baseline1['pred_p'])

            baselineR_gaussian.add_example(data['p'], r, 1.0 / cats_transformer.continuous_range)
            baselineR_clopper_pearson.add_example(data['p'], r, 1.0 / cats_transformer.continuous_range)
            
            evts += 1


    if log_fp.endswith('.gz'):
        len_text = ds_parse.update_progress(i+1)
    else:
        len_text = ds_parse.update_progress(bytes_count,tot_bytes)

    print('\nProcessed {} events out of {} lines'.format(evts,i+1))

    print('online_ips:',online_ips.get())

    print('baseline1_ips:', baseline1_ips.get())
    print('baseline1 gaussian ci:', baseline1_gaussian.get())
    print('baseline1 clopper pearson ci:', baseline1_clopper_pearson.get())

    print('baselineR_ips:',baselineR_ips.get())
    print('baselineR gaussian ci:', baselineR_gaussian.get())
    print('baselineR clopper pearson ci:', baselineR_clopper_pearson.get())


    print('online_snips:',online_snips.get())
    print('baseline1_snips:',baseline1_snips.get())
    print('baselineR_snips:',baselineR_snips.get())

    print('online_mle:',online_mle.get())
    print('baseline1_mle:',baseline1_mle.get())
    print('baselineR_mle:',baselineR_mle.get())

    print('online_cressieread:',online_cressieread.get())
    print('baseline1_cressieread:',baseline1_cressieread.get())
    print('baselineR_cressieread:',baselineR_cressieread.get())

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('-l','--log_fp', help="data file path (.json or .json.gz format - each line is a dsjson)", required=True)
    parser = cats_utils.set_custom_args(parser)
    args = parser.parse_args()
    cats_transformer = cats_utils.get_cats_transformer(args)

    compute_estimates(args.log_fp, cats_transformer)



# Empirical likehood estimator

from math import fsum, inf
from estimators.bandits import base
from typing import Optional


class Estimator(base.Estimator):
    # NB: This works better you use the true wmin and wmax
    #     which is _not_ the empirical minimum and maximum
    #     but rather the actual smallest and largest possible values
    def __init__(self, wmin: float = 0, wmax: float = inf):
        assert wmin < 1
        assert wmax > 1

        self.wmin = wmin
        self.wmax = wmax

        self.data = []

    def add_example(self, p_log: float, r: float, p_pred: float) -> None:
        w = p_pred / p_log
        assert w >= 0, 'Error: negative importance weight'

        self.data.append((1, w, r))
        self.wmax = max(self.wmax, w)
        self.wmin = min(self.wmin, w)

    def graddualobjective(self, n, beta):
        return fsum(c * (w - 1) / ((w - 1) * beta + n)
                    for c, w, _ in self.data)

    def get(self) -> Optional[float]:
        from scipy.optimize import brentq

        n = fsum(c for c, _, _ in self.data)
        if n == 0:
            return None

        betaub = n / (1 - self.wmin)
        betamax = min(betaub,
                      min(((n - c) / (1 - w)
                           for c, w, _ in self.data
                           if w < 1
                           ),
                          default=betaub))

        betalb = 0 if self.wmax == inf else n / (1 - self.wmax)
        betamin = max(betalb,
                      max(((n - c) / (1 - w)
                           for c, w, _ in self.data
                           if w > 1
                           ),
                          default=betalb))

        gradmin = self.graddualobjective(n, betamin)
        gradmax = self.graddualobjective(n, betamax)

        if gradmin * gradmax < 0:
            betastar = brentq(f=lambda x: self.graddualobjective(n, x),
                              a=betamin,
                              b=betamax)
        elif gradmin < 0:
            betastar = betamin
        else:
            betastar = betamax

        sumofw = fsum(c * w / ((w - 1) * betastar + n)
                      for c, w, _ in self.data)
        missing = max(0.0, 1.0 - sumofw)

        vhat = fsum(c * w * r / ((w - 1) * betastar + n)
                    for c, w, r in self.data)
        rhatmissing = fsum(c * r for c, _, r in self.data) / n
        vhat += missing * rhatmissing

        return vhat



from estimators.ccb import base
from typing import List, Optional


class Estimator(base.Estimator):
    def __init__(self, bandits_impl):
        self.impl = bandits_impl
        self.slots_count = 0

    def add_example(self, p_logs: List[float], rs: List[float], p_preds: List[float]) -> None:
        """Expects lists for logged probabilities, rewards and predicted probabilities. These should correspond to
        each slot. """

        if not isinstance(p_logs, list) and not isinstance(rs, list) and not isinstance(p_preds, list):
            raise ValueError('Error: p_logs, r and p_preds must be lists')

        if len(p_logs) != len(p_preds) and len(p_logs) != len(rs) and len(rs) != len(p_preds):
            raise ValueError(f'Error: p_logs, r and p_preds must be the same length, found {len(p_logs)}, {len(rs)} and'
                             f'{len(p_preds)} respectively')
        self.slots_count = max(self.slots_count, len(p_logs))
        self.impl.add_example(p_logs[0], rs[0], p_preds[0])

    def get(self) -> List[float]:
        if self.slots_count > 0:
            result = self.impl.get()
            return [result] + [0] * (self.slots_count - 1)
        return []


class Interval(base.Interval):
    def __init__(self, bandits_impl):
        self.impl = bandits_impl
        self.slots_count = 0

    def add_example(self, p_logs: List[float], rs: List[float], p_preds: List[float], p_drop: float = 0, n_drop: Optional[int] = None) -> None:
        """Expects lists for logged probabilities, rewards and predicted probabilities. These should correspond to
        each slot. """

        if not isinstance(p_logs, list) and not isinstance(rs, list) and not isinstance(p_preds, list):
            raise ValueError('Error: p_logs, r and p_preds must be lists')

        if len(p_logs) != len(p_preds) and len(p_logs) != len(rs) and len(rs) != len(p_preds):
            raise ValueError(f'Error: p_logs, r and p_preds must be the same length, found {len(p_logs)}, {len(rs)} and'
                             f'{len(p_preds)} respectively')
        self.slots_count = max(self.slots_count, len(p_logs))
        self.impl.add_example(p_logs[0], rs[0], p_preds[0], p_drop, n_drop)

    def get(self, alpha: float = 0.05) -> List[List[float]]:
        if self.slots_count > 0:
            result = self.impl.get()
            return [result] + [[0, 0]] * (self.slots_count - 1)
        return []



from scipy.stats import beta
from typing import List, Optional
from copy import deepcopy

class IncrementalFsum:
    """ Incremental version of https://en.wikipedia.org/wiki/Kahan_summation_algorithm """

    def __init__(self):
        self.partials = []

    def __iadd__(self, x):
        i = 0
        for y in self.partials:
            if abs(x) < abs(y):
                x, y = y, x
            hi = x + y
            lo = y - (hi - x)
            if lo:
                self.partials[i] = lo
                i += 1
            x = hi
        self.partials[i:] = [x]
        return self

    def __add__(self, other):
        result = IncrementalFsum()
        result.partials = deepcopy(self.partials)
        for y in other.partials:
            result += y
        return result

    def __float__(self):
        return sum(self.partials, 0.0)

def clopper_pearson(successes: float, n: float, alpha: float = 0.05) -> List[Optional[float]]:
    lower_bound = beta.ppf(alpha / 2, successes, n - successes + 1) if successes > 0 else 0
    upper_bound = beta.ppf(1 - alpha / 2, successes + 1, n - successes) if successes < n else 1
    return [lower_bound, upper_bound]



""" Interface for implementation of conditional contextual bandits estimators """

from abc import ABC, abstractmethod
from typing import List, Optional


class Estimator(ABC):
	""" Interface for implementation of conditional contextual bandits estimators """

	@abstractmethod
	def add_example(self, p_log: List, r: List, p_pred: List) -> None:
		""" 
		Args:
			p_log: List of probability of the logging policy
			r: List of reward for choosing an action in the given context
			p_pred: List of predicted probability of making decision
		"""
		...

	@abstractmethod
	def get(self) -> List[float]:
		""" Calculates the selected estimator

		Returns:
			The estimator value
		"""
		...


class Interval(ABC):
	""" Interface for implementation of conditional contextual bandits estimators interval """

	@abstractmethod
	def add_example(self, p_log: List[float], r: List[float], p_pred: List[float], p_drop: float = 0, n_drop: Optional[int] = None) -> None:
		""" 
		Args:
			p_log: List of probability of the logging policy
			r: List of reward for choosing an action in the given context
			p_pred: List of predicted probability of making decision
			p_drop: probability for event to be dropped
			n_drop: amount of dropped events between current and previous ones (populated as p_drop/(1-p_drop) if None)
		"""
		...

	@abstractmethod
	def get(self, alpha: float) -> List[List[float]]:
		""" Calculates the CI
		Args:
			alpha: alpha value
		Returns:
			Returns the confidence interval as list[float]
		"""
		...



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="vw-estimators",
    version="0.2.1",
    description="Python package of estimators to perform off-policy evaluation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/VowpalWabbit/estimators.git",
    license="BSD 3-Clause License",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering"
    ],
    packages=["estimators", "estimators.bandits", "estimators.ccb", "estimators.slates", "estimators.utils"],
    install_requires= ['scipy>=0.9'],
    tests_require=['pytest'],
    python_requires=">=3.6",
)




from estimators.bandits import base
from typing import List, Optional
from estimators.math import IncrementalFsum


def _logwealth(s, v: float, rho: float) -> float:
    from math import log

    def loggammalowerinc(*, a, x):
        import scipy.special as sc

        return log(sc.gammainc(a, x)) + sc.loggamma(a)
                        
    assert s + v + rho > 0
    assert rho > 0

    return (s + v
            + rho * log(rho)
            - (v + rho) * log(s + v + rho)
            + loggammalowerinc(a = v + rho, x = s + v + rho)
            - loggammalowerinc(a = rho, x = rho)
    )

def _lblogwealth(t: int, sumXt: float, v: float, rho: float, alpha: float) -> float:
    from math import log
    import scipy.optimize as so

    assert 0 < alpha < 1, alpha
    thres = -log(alpha)

    minmu = 0
    logwealthminmu = _logwealth(s=sumXt, v=v, rho=rho)

    if logwealthminmu <= thres:
        return minmu
    
    maxmu = min(1, sumXt/t)
    logwealthmaxmu = _logwealth(s=sumXt - t * maxmu, v=v, rho=rho)

    if logwealthmaxmu >= thres:
        return maxmu

    res = so.root_scalar(f = lambda mu: _logwealth(s=sumXt - t * mu, v=v, rho=rho) - thres,
                            method = 'brentq',
                            bracket = [ minmu, maxmu ])
    assert res.converged, res
    return res.root


class IntervalImpl(object):
    def __init__(self, rmin=0, rmax=1, adjust=True):
        super().__init__()
        
        assert rmin <= rmax, (rmin, rmax)
                
        self.rho = 1
        self.rmin = rmin
        self.rmax = rmax
        self.adjust = adjust
        
        self.t = 0

        self.sumwsqrsq = IncrementalFsum()
        self.sumwsqr = IncrementalFsum()
        self.sumwsq = IncrementalFsum()
        self.sumwr = IncrementalFsum()
        self.sumw = IncrementalFsum()
        self.sumwrxhatlow = IncrementalFsum()
        self.sumwxhatlow = IncrementalFsum()
        self.sumxhatlowsq = IncrementalFsum()
        self.sumwrxhathigh = IncrementalFsum()
        self.sumwxhathigh = IncrementalFsum()
        self.sumxhathighsq = IncrementalFsum()
            
    def add(self, w: float, r: float, p_drop: float, n_drop: int) -> None:
        assert w >= 0
        assert 0 <= p_drop < 1
        assert n_drop is None or n_drop >= 0
        
        if not self.adjust:
            r = min(self.rmax, max(self.rmin, r))
        else:
            self.rmin = min(self.rmin, r)
            self.rmax = max(self.rmax, r)
            
        if n_drop is None:
            n_drop = p_drop / (1 - p_drop)
            
        if n_drop > 0:
            import scipy.special as sc
            
            # we have to simulate presenting n_drop events with w=0 in a row, which we can do in closed form
            # Sum[(a/(b + s))^2, { s, 0, n - 1 }] 
            # a^2 PolyGamma[1,b]-a^2 PolyGamma[1,b+n] 
            
            sumXlow = (float(self.sumwr) - float(self.sumw) * self.rmin) / (self.rmax - self.rmin)
            alow = sumXlow + 1/2
            blow = self.t + 1
            self.sumxhatlowsq += alow**2 * (sc.polygamma(1, blow).item() - sc.polygamma(1, blow + n_drop).item())
            
            sumXhigh = (float(self.sumw) * self.rmax - float(self.sumwr)) / (self.rmax - self.rmin)
            ahigh = sumXhigh + 1/2
            bhigh = self.t + 1
            self.sumxhathighsq += ahigh**2 * (sc.polygamma(1, bhigh).item() - sc.polygamma(1, bhigh + n_drop).item())
            
            self.t += n_drop
            
        sumXlow = (float(self.sumwr) - float(self.sumw) * self.rmin) / (self.rmax - self.rmin)
        Xhatlow = (sumXlow + 1/2) / (self.t + 1)
        sumXhigh = (float(self.sumw) * self.rmax - float(self.sumwr)) / (self.rmax - self.rmin)
        Xhathigh = (sumXhigh + 1/2) / (self.t + 1)
        
        w /= (1 - p_drop)
        
        self.sumwsqrsq += (w * r)**2
        self.sumwsqr += w**2 * r
        self.sumwsq += w**2
        self.sumwr += w * r
        self.sumw += w
        self.sumwrxhatlow += w * r * Xhatlow
        self.sumwxhatlow += w * Xhatlow
        self.sumxhatlowsq += Xhatlow**2
        self.sumwrxhathigh += w * r * Xhathigh
        self.sumwxhathigh += w * Xhathigh
        self.sumxhathighsq += Xhathigh**2
            
        self.t += 1
        
    def get(self, alpha: float) -> List[Optional[float]]:
        if self.t == 0 or self.rmin == self.rmax:
            return [None, None]
        
        sumvlow = (  (float(self.sumwsqrsq) - 2 * self.rmin * float(self.sumwsqr) + self.rmin**2 * float(self.sumwsq)) / (self.rmax - self.rmin)**2
                   - 2 * (float(self.sumwrxhatlow) - self.rmin * float(self.sumwxhatlow)) / (self.rmax - self.rmin)
                   + float(self.sumxhatlowsq)
                  )
        sumXlow = (float(self.sumwr) - float(self.sumw) * self.rmin) / (self.rmax - self.rmin)
        l = _lblogwealth(t=self.t, sumXt=sumXlow, v=sumvlow, rho=self.rho, alpha=alpha/2)
        
        sumvhigh = (  (float(self.sumwsqrsq) - 2 * self.rmax * float(self.sumwsqr) + self.rmax**2 * float(self.sumwsq)) / (self.rmax - self.rmin)**2
                    + 2 * (float(self.sumwrxhathigh) - self.rmax * float(self.sumwxhathigh)) / (self.rmax - self.rmin)
                    + float(self.sumxhathighsq)
                   )
        sumXhigh = (float(self.sumw) * self.rmax - float(self.sumwr)) / (self.rmax - self.rmin)
        u = 1 - _lblogwealth(t=self.t, sumXt=sumXhigh, v=sumvhigh, rho=self.rho, alpha=alpha/2)
        
        return [self.rmin + l * (self.rmax - self.rmin), self.rmin + u * (self.rmax - self.rmin)]


class Interval(base.Interval):
    _impl: IntervalImpl

    def __init__(self, rmin: float = 0, rmax: float = 1, empirical_r_bounds = False):
        self._impl = IntervalImpl(rmin=rmin, rmax=rmax, adjust=empirical_r_bounds)

    def add_example(self, p_log: float, r: float, p_pred: float, p_drop: float = 0, n_drop: Optional[int] = None) -> None:
        self._impl.add(p_pred / p_log, r, p_drop, n_drop)

    def get(self, alpha: float = 0.05, atol: float = 1e-9) -> List[Optional[float]]:
        return self._impl.get(alpha)

    def __add__(self, other: 'Interval') -> 'Interval':
        raise "Not supported yet"
